"""Tests for the misc actors"""

import os
import time

from nose.plugins.attrib import attr

from tornado import testing

from kingpin.actors import exceptions
from kingpin.actors import misc


__author__ = "Mikhail Simin <mikhail@nextdoor.com>"


class IntegrationGenericHTTP(testing.AsyncTestCase):

    integration = True

    @attr("http", "integration")
    @testing.gen_test(timeout=60)
    def integration_get(self):
        actor = misc.GenericHTTP("Test", {"url": "http://httpbin.org/get"})

        yield actor.execute()

    @attr("http", "integration")
    @testing.gen_test(timeout=60)
    def integration_post(self):
        actor = misc.GenericHTTP(
            "Test", {"url": "http://httpbin.org/post", "data": {"foo": "bar"}}
        )

        yield actor.execute()

    @attr("http", "integration")
    @testing.gen_test(timeout=60)
    def integration_auth(self):
        actor = misc.GenericHTTP(
            "Test",
            {
                "url": "http://httpbin.org/basic-auth/unit/test",
                "username": "unit",
                "password": "test",
            },
        )

        yield actor.execute()

    @attr("http", "integration")
    @testing.gen_test(timeout=60)
    def integration_auth_fail(self):
        actor = misc.GenericHTTP(
            "Test",
            {
                "url": "http://httpbin.org/basic-auth/unit/test",
                "username": "wrong",
                "password": "pass",
            },
        )

        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor.execute()


class IntegrationMacro(testing.AsyncTestCase):

    integration = True

    @attr("http", "integration")
    @testing.gen_test
    def integration_execute(self):
        actor = misc.Macro(
            "Test", {"macro": "examples/test/sleep.json", "tokens": dict(os.environ)}
        )

        start = time.time()
        yield actor.execute()
        end = time.time()
        runtime = end - start
        self.assertTrue(runtime > 0.1)

    @attr("http", "integration")
    @testing.gen_test
    def integration_fail_without_env(self):
        # Actor should fail if tokens aren't passed for env. variables.
        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            misc.Macro("Test", {"macro": "examples/test/sleep.json"}, init_tokens={})

    @attr("http", "integration")
    @testing.gen_test
    def integration_execute_remote(self):
        gh_src = "https://raw.githubusercontent.com/Nextdoor/kingpin/master"
        # Successful __init__ on this actor validates downloading and parsing.
        misc.Macro(
            "Test",
            {"macro": gh_src + "/examples/test/sleep.json", "tokens": dict(os.environ)},
        )

    @attr("http", "integration", "dry")
    @testing.gen_test
    def integration_execute_dry(self):
        actor = misc.Macro(
            "Test",
            {"macro": "examples/test/sleep.json", "tokens": dict(os.environ)},
            dry=True,
        )

        start = time.time()
        yield actor.execute()
        end = time.time()
        runtime = end - start
        self.assertTrue(runtime < 0.1)



"""Simple integration tests for the AWS S3 actors."""

from nose.plugins.attrib import attr
import logging

from tornado import testing

# from kingpin import utils
# from kingpin.actors import exceptions
from kingpin.actors.aws import s3

__author__ = "Matt Wise <matt@nextdoor.com>"

log = logging.getLogger(__name__)
logging.getLogger("boto").setLevel(logging.INFO)


class IntegrationS3(testing.AsyncTestCase):

    """High level S3 Actor testing.

    These tests will check two things:
    * Creation of S3 buckets with a variety of options
    * Updating these S3 buckets with new parameters
    * Destroying the S3 buckets

    Requirements:
        You have to create an S3 Bucket named kingpin-integration-test and
        place it in the specified region (default us-east-1).

        You must have an AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY if using
        an IAM user/role and also an AWS_SESSION_TOKEN if using temporary access credentials,
        with permissions to read S3 bucket information.

    Note, these tests must be run in-order. The order is defined by
    their definition order in this file. Nose follows this order according
    to its documentation:

        http://nose.readthedocs.org/en/latest/writing_tests.html
    """

    integration = True

    bucket_name = "kingpin-integration-test"
    region = "us-east-1"

    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_01_create_bucket(self):
        actor = s3.Bucket(
            options={
                "name": self.bucket_name,
                "region": self.region,
                "state": "present",
                "logging": {"target": ""},
                "versioning": False,
            }
        )
        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_02a_set_bucket_policy(self):
        actor = s3.Bucket(
            options={
                "name": self.bucket_name,
                "region": self.region,
                "state": "present",
                "policy": "examples/aws.s3/amazon_put.json",
            }
        )
        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_02b_delete_bucket_policy(self):
        actor = s3.Bucket(
            options={
                "name": self.bucket_name,
                "region": self.region,
                "state": "present",
                "policy": "",
            }
        )
        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_03a_enable_versioning(self):
        actor = s3.Bucket(
            options={
                "name": self.bucket_name,
                "region": self.region,
                "state": "present",
                "versioning": True,
            }
        )
        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_03b_disable_versioning(self):
        actor = s3.Bucket(
            options={
                "name": self.bucket_name,
                "region": self.region,
                "state": "present",
                "versioning": False,
            }
        )
        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_04a_enable_lifecycle_management(self):
        actor = s3.Bucket(
            options={
                "name": self.bucket_name,
                "region": self.region,
                "state": "present",
                "lifecycle": [
                    {
                        "id": "test",
                        "prefix": "/",
                        "status": "Enabled",
                        "expiration": 30,
                        "transition": {"days": 10, "storage_class": "GLACIER"},
                    }
                ],
            }
        )
        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_04b_update_lifecycle_management(self):
        actor = s3.Bucket(
            options={
                "name": self.bucket_name,
                "region": self.region,
                "state": "present",
                "lifecycle": [
                    {
                        "id": "test",
                        "prefix": "/",
                        "status": "Enabled",
                        "expiration": 180,
                        "transition": {"days": 90, "storage_class": "STANDARD_IA"},
                    }
                ],
            }
        )
        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_04c_disable_lifecycle_management(self):
        actor = s3.Bucket(
            options={
                "name": self.bucket_name,
                "region": self.region,
                "state": "present",
                "lifecycle": [],
            }
        )
        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr("aws", "integration")
    @testing.gen_test(timeout=60)
    def integration_09_delete_bucket(self):
        actor = s3.Bucket(
            options={
                "name": self.bucket_name,
                "region": self.region,
                "state": "absent",
            }
        )
        done = yield actor.execute()
        self.assertEqual(done, None)



import copy
import mock
import logging
import json

from tornado import testing
from tornado import httpclient

from kingpin.actors import exceptions
from kingpin.actors import spotinst
from kingpin.actors.test.helper import mock_tornado, tornado_value

__author__ = "Matt Wise <matt@nextdoor.com>"


class TestSpotinstException(testing.AsyncTestCase):
    def test_no_json_in_body(self):
        fake_body = "400 Bad Mmmkay"
        exc = spotinst.SpotinstException(fake_body)
        self.assertEqual("Unknown error: 400 Bad Mmmkay", str(exc))

    def test_invalid_auth_response(self):
        fake_resp_body = mock.MagicMock(name="response_body")
        fake_resp_body.body = json.dumps(
            {
                "request": {
                    "id": "fake_id",
                    "url": "/fake",
                    "method": "GET",
                    "timestamp": "2016-12-28T22:36:36.324Z",
                },
                "response": {"error": "invalid auth", "error_id": "NOAUTH"},
            }
        )
        source_exc = httpclient.HTTPError(400, "400 Bad Request", fake_resp_body)
        fake_exc = spotinst.SpotinstException(source_exc)
        self.assertEqual(
            "Spotinst Request ID (fake_id) GET /fake: invalid auth", str(fake_exc)
        )

    def test_group_validation_errors(self):
        fake_resp_body = mock.MagicMock(name="response_body")
        fake_resp_body.body = json.dumps(
            {
                "request": {
                    "id": "fake_id",
                    "url": "/fake",
                    "method": "GET",
                    "timestamp": "2016-12-28T22:36:36.324Z",
                },
                "response": {
                    "errors": [
                        {
                            "message": "Cant create spot requests.",
                            "code": "GENERAL_ERROR",
                        },
                        {
                            "message": "AMI ami-16fc4976 with an...",
                            "code": "UnsupportedOperation",
                        },
                    ]
                },
            }
        )
        source_exc = httpclient.HTTPError(400, "400 Bad Request", fake_resp_body)
        fake_exc = spotinst.SpotinstException(source_exc)
        self.assertEqual(
            (
                "Spotinst Request ID (fake_id) GET /fake: GENERAL_ERROR: Cant "
                "create spot requests., UnsupportedOperation: AMI ami-16fc4976 "
                "with an..."
            ),
            str(fake_exc),
        )

    def test_unknown_error_body(self):
        fake_resp_body = mock.MagicMock(name="response_body")
        fake_resp_body.body = json.dumps(
            {
                "request": {
                    "id": "fake_id",
                    "url": "/fake",
                    "method": "GET",
                    "timestamp": "2016-12-28T22:36:36.324Z",
                },
                "response": {"something": "else"},
            }
        )
        source_exc = httpclient.HTTPError(400, "400 Bad Request", fake_resp_body)
        fake_exc = spotinst.SpotinstException(source_exc)
        self.assertEqual(
            ("Spotinst Request ID (fake_id) GET /fake: " "{'something': 'else'}"),
            str(fake_exc),
        )


class TestSpotinstBase(testing.AsyncTestCase):

    """Unit tests for the packagecloud Base actor."""

    def setUp(self, *args, **kwargs):
        super(TestSpotinstBase, self).setUp(*args, **kwargs)
        spotinst.TOKEN = "Unittest"
        spotinst.DEBUG = True
        spotinst.ACCOUNT_ID = "act-test"

    def test_init_with_debug_disabled(self):
        spotinst.DEBUG = False
        spotinst.SpotinstBase("Unit Test Action", {})
        self.assertEqual(20, logging.getLogger("tornado_rest_client.api").level)

    def test_init_missing_token(self):
        spotinst.TOKEN = None
        with self.assertRaises(exceptions.InvalidCredentials):
            spotinst.SpotinstBase("Unit Test Action", {})

    def test_init_without_account_id(self):
        spotinst.ACCOUNT_ID = None
        with self.assertRaises(exceptions.InvalidCredentials):
            spotinst.SpotinstBase("Unit Test Action", {})


class TestElastiGroup(testing.AsyncTestCase):

    """Unit tests for the ElastiGroup actor."""

    def setUp(self, *args, **kwargs):
        super(TestElastiGroup, self).setUp(*args, **kwargs)
        file = "examples/test/spotinst.elastigroup/unittest.json"
        spotinst.TOKEN = "Unittest"
        spotinst.ACCOUNT_ID = "act-test"

        # Manually inject some fake values for the subnet/secgrp/zone
        init_tokens = {
            "SECGRP": "sg-123123",
            "ZONE": "us-test-1a",
            "SUBNET": "sn-123123",
        }

        self.actor = spotinst.ElastiGroup(
            "unittest",
            {
                "name": "unittest",
                "config": file,
                "wait_on_create": True,
                "wait_on_roll": True,
            },
            init_tokens=init_tokens,
        )
        self.actor._client = mock.Mock()

    def test_init_with_string_roll_settings(self):
        with self.assertRaises(exceptions.InvalidOptions):
            self.actor = spotinst.ElastiGroup(
                options={
                    "name": "unittest",
                    "config": "junk",
                    "roll_batch_size": "some_number",
                }
            )

    def test_parse_group_config(self):
        self.assertEqual(
            (self.actor._config["group"]["compute"]["availabilityZones"][0]["name"]),
            "us-test-1a",
        )
        self.assertEqual(self.actor._config["group"]["name"], "unittest")

    def test_parse_group_config_no_config(self):
        self.actor._options["config"] = None
        self.assertEqual(None, self.actor._parse_group_config())

    def test_parse_group_config_missing_token(self):
        del self.actor._init_tokens["ZONE"]
        with self.assertRaises(exceptions.InvalidOptions):
            self.actor._parse_group_config()

    @testing.gen_test
    def test_list_groups(self):
        list_of_groups = {
            "request": {
                "id": "fake_id",
                "url": "/fake",
                "method": "GET",
                "timestamp": "2016-12-28T22:36:36.324Z",
            },
            "response": {"items": [{"group": {"name": "test"}}]},
        }

        self.actor._client.aws.ec2.list_groups.http_get = mock_tornado(list_of_groups)
        ret = yield self.actor._list_groups()
        self.assertEqual(ret, [{"group": {"name": "test"}}])

    @testing.gen_test
    def test_get_group(self):
        matching_group = {
            "name": "unittest",
            "id": "bogus",
        }
        self.actor._list_groups = mock_tornado([matching_group])

        ret = yield self.actor._get_group()
        self.assertEqual(ret, {"group": matching_group})

    @testing.gen_test
    def test_get_group_too_many_results(self):
        matching_group = {
            "name": "unittest",
            "id": "bogus",
        }
        self.actor._list_groups = mock_tornado([matching_group, matching_group])

        with self.assertRaises(exceptions.InvalidOptions):
            yield self.actor._get_group()

    @testing.gen_test
    def test_get_group_no_groups(self):
        self.actor._list_groups = mock_tornado(None)
        ret = yield self.actor._get_group()
        self.assertEqual(ret, None)

    @testing.gen_test
    def test_get_group_no_matching_groups(self):
        unmatching_group = {
            "name": "unittest-not-matching",
            "id": "bogus",
        }
        self.actor._list_groups = mock_tornado([unmatching_group])
        ret = yield self.actor._get_group()
        self.assertEqual(ret, None)

    @testing.gen_test
    def test_precache(self):
        fake_group = {
            "name": "unittest",
            "id": "bogus",
            "group": {"capacity": {"target": 128}},
        }
        self.actor._get_group = mock_tornado(fake_group)
        self.actor._validate_group = mock_tornado(None)
        yield self.actor._precache()

        # First make sure we stored the group
        self.assertEqual(fake_group, self.actor._group)

        # Second, make sure we overwrote the user config's [capacity][target]
        # setting with the spotinst value
        self.assertEqual(self.actor._config["group"]["capacity"]["target"], 128)

    @testing.gen_test
    def test_validate_group(self):
        fake_ret = {"ok": True}
        mock_client = mock.MagicMock()
        mock_client.http_post.return_value = tornado_value(fake_ret)
        self.actor._client.aws.ec2.validate_group = mock_client

        yield self.actor._validate_group()
        mock_client.http_post.assert_called_with(group=self.actor._config["group"])

    @testing.gen_test
    def test_get_state(self):
        self.actor._group = True
        ret = yield self.actor._get_state()
        self.assertEqual("present", ret)

    @testing.gen_test
    def test_get_state_false(self):
        self.actor._group = None
        ret = yield self.actor._get_state()
        self.assertEqual("absent", ret)

    @testing.gen_test
    def test_set_state_present(self):
        fake_ret = {"ok": True, "response": {"items": [{"desc": "new group"}]}}
        mock_client = mock.MagicMock()
        mock_client.http_post.return_value = tornado_value(fake_ret)
        self.actor._client.aws.ec2.create_group = mock_client

        # Mock out the _get_group method which is called again after the
        # create_group API call is made.
        self.actor._get_group = mock_tornado(None)

        # Mock out the call to our wait_until_stable() call since thats not in
        # the scope of this test.
        self.actor._wait_until_stable = mock_tornado(None)

        yield self.actor._set_state()
        mock_client.http_post.assert_called_with(group=self.actor._config["group"])

    @testing.gen_test
    def test_set_state_absent(self):
        # First, lets copy the desired configuration blob. The first test,
        # we'll copy the blob and we'll ensure that they are the same.
        self.actor._group = copy.deepcopy(self.actor._config)

        # Insert some fake data that would normally have been returned in the
        # included blob from Spotinst.
        self.actor._group["group"]["id"] = "sig-1234123"
        self.actor._group["group"]["createdAt"] = "timestamp"
        self.actor._group["group"]["updatedAt"] = "timestamp"

        fake_ret = {"ok": True}
        self.actor._options["state"] = "absent"
        mock_client = mock.MagicMock()
        mock_client.http_delete.return_value = tornado_value(fake_ret)
        self.actor._client.aws.ec2.delete_group.return_value = mock_client

        yield self.actor._set_state()
        mock_client.http_delete.assert_called_with()

    @testing.gen_test
    def test_compare_config(self):
        # First, lets copy the desired configuration blob. The first test,
        # we'll copy the blob and we'll ensure that they are the same.
        self.actor._group = copy.deepcopy(self.actor._config)

        # Insert some fake data that would normally have been returned in the
        # included blob from Spotinst.
        self.actor._group["group"]["id"] = "sig-1234123"
        self.actor._group["group"]["createdAt"] = "timestamp"
        self.actor._group["group"]["updatedAt"] = "timestamp"

        # This should return True because the configs are identical..
        ret = yield self.actor._compare_config()
        self.assertEqual(True, ret)

        # Now, lets modify the ElastiGroup config a bit.. the diff should
        # return false.
        self.actor._group["group"]["description"] = "new description"
        ret = yield self.actor._compare_config()
        self.assertEqual(False, ret)

    @testing.gen_test
    def test_compare_config_not_existing(self):
        # Pretend that the group doesn't exist at all in Spotinst
        self.actor._group = None

        # This should return True because the config simply doesnt exist in
        # Spotinst. The _set_state() method will create it during a real run.
        ret = yield self.actor._compare_config()
        self.assertEqual(True, ret)

    @testing.gen_test
    def test_get_config(self):
        self.actor._group = 1
        ret = yield self.actor._get_config()
        self.assertEqual(1, ret)

    @testing.gen_test
    def test_set_config(self):
        # First, lets copy the desired configuration blob. The first test,
        # we'll copy the blob and we'll ensure that they are the same.
        self.actor._group = copy.deepcopy(self.actor._config)

        # Insert some fake data that would normally have been returned in the
        # included blob from Spotinst.
        self.actor._group["group"]["id"] = "sig-1234123"
        self.actor._group["group"]["createdAt"] = "timestamp"
        self.actor._group["group"]["updatedAt"] = "timestamp"

        # Pretend to roll the group if a change is made
        self.actor._options["roll_on_change"] = True
        self.actor._roll_group = mock.MagicMock()
        self.actor._roll_group.return_value = tornado_value(None)

        # Mock out the update_group call..
        fake_ret = {"response": {"items": [{"group": "object"}]}}
        mock_client = mock.MagicMock()
        mock_client.http_put.return_value = tornado_value(fake_ret)
        self.actor._client.aws.ec2.update_group.return_value = mock_client

        yield self.actor._set_config()

        mock_client.http_put.assert_called_with(group=self.actor._config["group"])
        self.assertEqual(self.actor._group["group"], {"group": "object"})
        self.actor._roll_group.assert_has_calls([mock.call])

    @testing.gen_test
    def test_roll_group(self):
        # First, lets copy the desired configuration blob. The first test,
        # we'll copy the blob and we'll ensure that they are the same.
        self.actor._group = copy.deepcopy(self.actor._config)

        # Insert some fake data that would normally have been returned in the
        # included blob from Spotinst.
        self.actor._group["group"]["id"] = "sig-1234123"
        self.actor._group["group"]["createdAt"] = "timestamp"
        self.actor._group["group"]["updatedAt"] = "timestamp"

        # Mock out the returned calls from Spotinst. The first time we call the
        # roll status method, we'll return no rolls in progress. The second
        # time, there will be a single roll in progress (pretending that the
        # roll_group call was successful), and then the third time it will
        # return no rolls in progress again (pretending that we're done).
        in_progress = {
            "id": "sbgd-44e6d801",
            "status": "in_progress",
            "progress": {"unit": "percent", "value": 0},
            "createdAt": "2017-01-05T15:48:28.000+0000",
            "updatedAt": "2017-01-05T15:49:15.000+0000",
        }
        finished = {
            "id": "sbgd-9f1aa4f6",
            "status": "finished",
            "progress": {"unit": "percent", "value": 100},
            "createdAt": "2017-01-05T15:06:25.000+0000",
            "updatedAt": "2017-01-05T15:22:17.000+0000",
        }

        roll_responses = [
            # First response, no rolls are in progress
            tornado_value(
                {
                    "request": {
                        "id": "request-1-no-in-progress",
                        "url": "/aws/ec2/group/sig-b55014f1/roll?limit=5",
                        "method": "GET",
                        "timestamp": "2017-01-05T15:50:44.215Z",
                    },
                    "response": {
                        "status": {"code": 200, "message": "OK"},
                        "kind": "spotinst:aws:ec2:group:roll",
                        "items": [finished],
                    },
                }
            ),
            # Ok now pretend that one roll is in progress, and one is finished
            tornado_value(
                {
                    "request": {
                        "id": "request-2-one-in-progress",
                        "url": "/aws/ec2/group/sig-b55014f1/roll?limit=5",
                        "method": "GET",
                        "timestamp": "2017-01-05T15:50:44.215Z",
                    },
                    "response": {
                        "status": {"code": 200, "message": "OK"},
                        "kind": "spotinst:aws:ec2:group:roll",
                        "items": [in_progress, finished],
                    },
                }
            ),
            # Finally, no rolls in progress.
            tornado_value(
                {
                    "request": {
                        "id": "request-3-none-in-progress",
                        "url": "/aws/ec2/group/sig-b55014f1/roll?limit=5",
                        "method": "GET",
                        "timestamp": "2017-01-05T15:50:44.215Z",
                    },
                    "response": {
                        "status": {"code": 200, "message": "OK"},
                        "kind": "spotinst:aws:ec2:group:roll",
                        "items": [finished, finished],
                    },
                }
            ),
        ]

        # Generate a basic mock object for the entire 'roll' RestConsumer. Mock
        # out the http_get() method to return back the three fake responses
        # above. Mock out the http_put() method to just return safely.
        roll_mock = mock.MagicMock()
        roll_mock.http_get.side_effect = roll_responses
        roll_mock.http_put.side_effect = [tornado_value(None)]
        self.actor._client.aws.ec2.roll.return_value = roll_mock

        # Make the call, and make sure we wait for all roll operations to
        # finish
        self.actor._options["wait_on_roll"] = True
        yield self.actor._roll_group(delay=0.01)

        # Now verify that all the calls were made to the roll_mock
        roll_mock.assert_has_calls(
            [
                mock.call.http_get(),
                mock.call.http_put(gracePeriod=600, batchSizePercentage=20),
                mock.call.http_get(),
                mock.call.http_get(),
            ]
        )

    @testing.gen_test
    def test_wait_until_stable(self):
        # First, lets copy the desired configuration blob. The first test,
        # we'll copy the blob and we'll ensure that they are the same.
        self.actor._group = copy.deepcopy(self.actor._config)

        # Insert some fake data that would normally have been returned in the
        # included blob from Spotinst.
        self.actor._group["group"]["id"] = "sig-1234123"
        self.actor._group["group"]["createdAt"] = "timestamp"
        self.actor._group["group"]["updatedAt"] = "timestamp"

        # Mock out what a pending vs fulfilled instance looks like
        pending = {
            "spotInstanceRequestId": "sir-n8688grq",
            "instanceId": None,
            "instanceType": "t1.micro",
            "product": "Linux/UNIX (Amazon VPC)",
            "availabilityZone": "us-west-2a",
            "createdAt": "2017-01-03T22:30:56.000Z",
            "status": "pending-evaluation",
        }
        fullfilled = {
            "spotInstanceRequestId": "sir-n8688grq",
            "instanceId": "i-abcdefg",
            "instanceType": "t1.micro",
            "product": "Linux/UNIX (Amazon VPC)",
            "availabilityZone": "us-west-2a",
            "createdAt": "2017-01-03T22:30:56.000Z",
            "status": "fullfilled",
        }

        # Create a mock for the group_status API call that returns different
        # results on the 3rd time its called. The first two times, the
        # instances will be in a pending-evaluation state, the third call they
        # will be in a fullfilled state.
        group_status_mock = mock.MagicMock("group_status")
        self.actor._client.aws.ec2.group_status().http_get = group_status_mock
        self.actor._client.aws.ec2.group_status().http_get.side_effect = [
            tornado_value({"response": {"items": [pending, pending]}}),
            tornado_value({"response": {"items": [pending, pending]}}),
            tornado_value({"response": {"items": [fullfilled, fullfilled]}}),
        ]

        # Now make the call
        yield self.actor._wait_until_stable(delay=0.01)
        group_status_mock.assert_has_calls([mock.call(), mock.call(), mock.call()])



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.aws.base`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The AWS Actors allow you to interact with the resources (such as SQS and ELB)
inside your Amazon AWS account. These actors all support dry runs properly, but
each actor has its own caveats with ``dry=True``. Please read the instructions
below for using each actor.

**Required Environment Variables**

_Note, these can be skipped only if you have a .aws/credentials file in place._

:AWS_ACCESS_KEY_ID:
  Your AWS access key

:AWS_SECRET_ACCESS_KEY:
  Your AWS secret

:AWS_SESSION_TOKEN:
  Your AWS session token
  Only needed if you are using temporary access credentials
"""

import logging

from boto3 import exceptions as boto3_exceptions
from botocore import exceptions as botocore_exceptions
from botocore import config as botocore_config
from tornado import concurrent
from tornado import gen
from tornado import ioloop
import boto3

from kingpin import utils
from kingpin import exceptions as kingpin_exceptions
from kingpin.actors import base
from kingpin.actors import exceptions
from kingpin.actors.aws import api_call_queue
from kingpin.actors.aws import settings as aws_settings

log = logging.getLogger(__name__)

__author__ = "Mikhail Simin <mikhail@nextdoor.com>"

EXECUTOR = concurrent.futures.ThreadPoolExecutor(10)

NAMED_API_CALL_QUEUES = {}


class ELBNotFound(exceptions.RecoverableActorFailure):
    """Raised when an ELB is not found"""


class InvalidMetaData(exceptions.UnrecoverableActorFailure):
    """Raised when fetching AWS metadata."""


class InvalidPolicy(exceptions.RecoverableActorFailure):
    """Raised when Amazon indicates that policy JSON is invalid."""


class AWSBaseActor(base.BaseActor):
    # Get references to existing objects that are used by the
    # tornado.concurrent.run_on_executor() decorator.
    ioloop = ioloop.IOLoop.current()
    executor = EXECUTOR

    all_options = {"region": (str, None, "AWS Region (or zone) to connect to.")}

    def __init__(self, *args, **kwargs):
        """Check for required settings."""

        super(AWSBaseActor, self).__init__(*args, **kwargs)

        # By default, we will try to let Boto handle discovering its
        # credentials at instantiation time. This _can_ result in synchronous
        # API calls to the Metadata service, but those should be fast.
        #
        # In the event though that someone has explicitly set the AWS access
        # keys in the environment (either for the purposes of a unit test, or
        # because they wanted to), we use those values.
        #
        # Note: these get defualted to None in aws_settings if they are not
        # found which will tell boto3 to fallback to default behavior.
        boto3_client_kwargs = {}
        boto3_client_kwargs["aws_access_key_id"] = aws_settings.AWS_ACCESS_KEY_ID
        boto3_client_kwargs[
            "aws_secret_access_key"
        ] = aws_settings.AWS_SECRET_ACCESS_KEY
        boto3_client_kwargs["aws_session_token"] = aws_settings.AWS_SESSION_TOKEN

        # Establish connection objects that don't require a region
        self.iam_conn = boto3.client(
            service_name="iam", config=None, **boto3_client_kwargs
        )

        # Establish region-specific connection objects.
        self.region = self.option("region")
        if not self.region:
            return

        # Generate our common config options that will be passed into the boto3
        # client constructors...
        boto_config = botocore_config.Config(
            region_name=self.region,
            retries={
                "max_attempts": aws_settings.AWS_MAX_ATTEMPTS,
                "mode": aws_settings.AWS_RETRY_MODE,
            },
        )

        self.ecs_conn = boto3.client(
            service_name="ecs", config=boto_config, **boto3_client_kwargs
        )
        self.cf3_conn = boto3.client(
            service_name="cloudformation", config=boto_config, **boto3_client_kwargs
        )
        self.sqs_conn = boto3.client(
            service_name="sqs", config=boto_config, **boto3_client_kwargs
        )
        self.s3_conn = boto3.client(
            service_name="s3", config=boto_config, **boto3_client_kwargs
        )

    @concurrent.run_on_executor
    @utils.exception_logger
    def api_call(self, api_function, *args, **kwargs):
        """Execute `api_function` in a concurrent thread.

        Example:
            >>> zones = yield thread(ec2_conn.get_all_zones)

        This allows execution of any function in a thread without having
        to write a wrapper method that is decorated with run_on_executor()
        """
        try:
            return api_function(*args, **kwargs)
        except boto3_exceptions.Boto3Error as e:
            raise self._wrap_boto_exception(e)

    @gen.coroutine
    @utils.exception_logger
    def api_call_with_queueing(self, api_function, queue_name, *args, **kwargs):
        """
        Execute `api_function` in a serialized queue.

        Concurrent calls to this function are serialized into a queue.
        When any api function hits rate throttling, it backs up exponentially.

        The retry loop will always have a pause between sequential calls,
        and the delay between the calls will increase as
        recoverable api failures happen.

        The api function is assumed to be a synchronous function.
        It will be run on a concurrent thread using run_on_executor.

        The queue_identifier argument specifies which queue to use.
        If the queue doesn't exist, it will be created.

        Example:
            >>> zones = yield api_call_with_queueing(
            >>>     ec2_conn.get_all_zones, queue_name='get_all_zones')
        """
        if queue_name not in NAMED_API_CALL_QUEUES:
            NAMED_API_CALL_QUEUES[queue_name] = api_call_queue.ApiCallQueue()
        queue = NAMED_API_CALL_QUEUES[queue_name]
        try:
            result = yield queue.call(api_function, *args, **kwargs)
        except botocore_exceptions.ClientError as e:
            raise self._wrap_boto_exception(e)
        else:
            raise gen.Return(result)

    def _wrap_boto_exception(self, e):
        if isinstance(e, boto3_exceptions.Boto3Error):
            return exceptions.RecoverableActorFailure("Boto3 had a failure: %s" % e)
        return e

    def _parse_policy_json(self, policy):
        """Parse a single JSON file into an Amazon policy.

        Validates that the policy document can be parsed, strips out any
        comments, and fills in any environmental tokens. Returns a dictionary
        of the contents.

        Returns None if the input is None.

        args:
            policy: The Policy JSON file to read.

        returns:
            A dictionary of the parsed policy.
        """
        if policy is None:
            return None

        # Run through any supplied Inline IAM Policies and verify that they're
        # not corrupt very early on.
        self.log.debug("Parsing and validating %s" % policy)

        try:
            p_doc = utils.convert_script_to_dict(
                script_file=policy, tokens=self._init_tokens
            )
        except kingpin_exceptions.InvalidScript as e:
            raise exceptions.UnrecoverableActorFailure(
                "Error parsing %s: %s" % (policy, e)
            )

        return p_doc


class EnsurableAWSBaseActor(AWSBaseActor, base.EnsurableBaseActor):

    """Ensurable version of the AWS Base Actor"""
